@@ -1,4 +1,4 @@
-use zookeeper::{ZooKeeper, WatchedEvent, WatchedEventType, ZkError, ZkResult};
+use zookeeper::{ZooKeeper, WatchedEvent, WatchedEventType, ZkError, ZkResult, ZooKeeperExt};
 use std::sync::{Arc, RwLock};
 use serde::{Serialize};
 use serde::de::DeserializeOwned;
@@ -9,6 +9,8 @@ use std::sync::RwLockReadGuard;
 use serde_json::Value;
 use crossbeam_channel::Receiver;
 use std::fmt::Debug;
+use rand::Rng;
+
 
 const MAX_TIMING_DELTA: i64 = 30000; // ms
 const LOCK_POLL_INTERVAL: u64 = 5; // ms
@@ -61,6 +63,8 @@ pub struct ZkState<T: Serialize + DeserializeOwned + Send + Sync> {
 }
 impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
     pub fn new(zk: Arc<ZooKeeper>, zk_path: String, initial_state: T) -> anyhow::Result<Self> {
+        zk.ensure_path(zk_path.as_str())?;
+
         let id = uuid::Uuid::new_v4().to_string();
 
         let (chan_tx, chan_rx) = crossbeam_channel::unbounded();
@@ -89,7 +93,7 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
     /// will block waiting for that to show up.
     ///
     /// TODO describe this better
-    pub fn expect(zk: Arc<ZooKeeper>, zk_path: String) -> anyhow::Result<Self> {
+    pub fn expect(zk: Arc<ZooKeeper>, zk_path: String, timeout: Option<Duration>) -> anyhow::Result<Self> {
         let id = uuid::Uuid::new_v4().to_string();
 
         // block, waiting for our initial state to show up in zookeeper
@@ -102,7 +106,10 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
         if let Ok(inner) = raw_data {
             data = inner.0;
         } else {
-            let _ = l_rx.recv();
+            match timeout {
+                None => l_rx.recv()?,
+                Some(d) => l_rx.recv_timeout(d)?
+            }
             data = zk.get_data(format!("{}/payload", &zk_path).as_str(), false)?.0;
         }
 
@@ -129,11 +136,8 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
         let path = format!("{}/payload", &self.state.read().unwrap().zk_path);
         // if the path doesn't exist, let's make it and populate it
         if self.zk.exists(path.as_str(), false).unwrap().is_none() {
-            log::debug!("{} does not exist, creating", &path);
-            self.zk.create(&self.state.read().unwrap().zk_path, vec![], zookeeper::Acl::open_unsafe().clone(), zookeeper::CreateMode::Persistent)?;
+            log::debug!(target: &*format!("zkstate::{}", &self.id), "{} does not exist, creating", &path);
 
-
-            // we need to populate it
             let data = self.inner.read().unwrap();
             let inner = serde_json::to_vec(&*data).unwrap();
             self.zk.create(path.as_str(), inner, zookeeper::Acl::open_unsafe().clone(), zookeeper::CreateMode::Persistent)?;
@@ -153,8 +157,9 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
             let zk = zk;
             let inner = inner;
             let state = state;
+            let mut rng = rand::thread_rng();
             loop {
-                thread::sleep(Duration::from_secs(5)); // TODO make this configurable
+                thread::sleep(Duration::from_secs_f32(rng.gen_range(3.0..7.0))); // TODO make this configurable
 
                 let handle = state.read().unwrap();
                 let path = format!("{}/payload", handle.zk_path);
@@ -206,17 +211,19 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
 
         // acquire write lock for the internal object
         let mut inner = self.inner.write().unwrap();
-        let state = self.state.write().unwrap();
+        let mut state = self.state.write().unwrap();
 
         // get write lock from zookeeper to prevent anyone from modifying this object while we're
         // committing it
-        let latch_path = format!("{}/write_lock", &state.zk_path);
-        let latch = zookeeper::recipes::leader::LeaderLatch::new(self.zk.clone(), self.id.clone(), latch_path);
+        let latch = zookeeper::recipes::leader::LeaderLatch::new(self.zk.clone(), self.id.clone(),  format!("{}/write_lock", &state.zk_path));
         latch.start()?;
 
         let mut total_time = 0;
         loop {
             if latch.has_leadership() { break; }
+
+            log::trace!("I do not have write_lock, waiting {}ms before trying again", LOCK_POLL_INTERVAL);
+
             thread::sleep(Duration::from_millis(LOCK_POLL_INTERVAL));
             if total_time > LOCK_POLL_TIMEOUT {
                 return Err(ZkStructError::LockAcquireTimeout)
@@ -225,17 +232,15 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
             }
         }
 
-        // pre change
+        // internal state, pre change
         let a = serde_json::to_value(&*inner).unwrap();
 
         // at this point, we should have an exclusive lock on the object so we execute the closure
         closure(&mut inner);
 
-        // post change
+        // internal state, post change
         let b = serde_json::to_value(&*inner).unwrap();
 
-        emit_updates(&a, &b, &state);
-
         let raw_data = serde_json::to_vec(&*inner).unwrap();
         let update_op = self.zk.set_data(path.as_str(), raw_data, Some(state.epoch));
         match update_op {
@@ -248,6 +253,12 @@ impl<T: Serialize + DeserializeOwned + Send + Sync + 'static> ZkState<T> {
             }
         }
 
+        // adjust the internal epoch to match the remote
+        let result_stats = update_op.unwrap();
+        state.epoch = result_stats.version;
+
+        emit_updates(&a, &b, &state);
+
         drop(inner); // drop the write lock on the inner object
         drop(state); // drop the write lock on the internal state object
 
@@ -404,6 +415,7 @@ fn emit_updates(a: &serde_json::Value, b: &serde_json::Value, state: &InternalSt
             },
         };
         if op != Change::Unchanged() {
+            log::trace!(target: &*format!("zkstate::{}", &state.id), "enqueuing operation {:?}", &op);
             let _insert = state.chan_tx.send(op);
         }
     }
